/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: SecurityCurrent.java 1121 2006-09-27 08:51:06Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.security.propagation.context;

import org.objectweb.easybeans.security.api.EZBSecurityContext;
import org.objectweb.easybeans.security.api.EZBSecurityCurrent;

/**
 * Manages the current security context associated to the current thread.
 * @author Florent Benoit
 */
public class SecurityCurrent implements EZBSecurityCurrent {

    /**
     * Inherited Local thread used to keep the security context.
     */
    private static InheritableThreadLocal<EZBSecurityContext> threadLocal;

    /**
     * Static Security Context that is applied on all threads (used for heavy
     * client).
     */
    private static EZBSecurityContext globalContext = null;

    /**
     * Default security context.
     */
    private static final EZBSecurityContext DEFAULT_CTX = new SecurityContext();

    /**
     * Init the thread
     */
    static {
        threadLocal = new InheritableThreadLocal<EZBSecurityContext>();
        threadLocal.set(new SecurityContext());
    }

    /**
     * Unique instance of this current object.
     */
    private static EZBSecurityCurrent unique = initCurrent();

    /**
     * Build the unique instance of the security current.<br>
     * It can use JOnAS current object.
     * @return unique instance.
     */
    @SuppressWarnings("unchecked")
    private static EZBSecurityCurrent initCurrent() {
        String externalPropertyClass = System.getProperty(SECURITY_CURRENT_PROPERTY);
        if (externalPropertyClass != null) {
            EZBSecurityCurrent current = null;
            // use this external current object
            Class<EZBSecurityCurrent> currentClass = null;
            try {
                currentClass = (Class<EZBSecurityCurrent>) Thread.currentThread().getContextClassLoader().loadClass(
                        externalPropertyClass);
            } catch (ClassNotFoundException e) {
                throw new IllegalStateException("Cannot find the class '" + externalPropertyClass + "'", e);
            } catch (ClassCastException e) {
                throw new IllegalStateException("'" + externalPropertyClass
                        + "' class is not an instance of EZBSecurityCurrent interface.", e);
            }

            try {
                current = currentClass.newInstance();
            } catch (InstantiationException e) {
                throw new IllegalStateException("Cannot create an instance of the class '" + externalPropertyClass + "'", e);
            } catch (IllegalAccessException e) {
                throw new IllegalStateException("Cannot create an instance of the class '" + externalPropertyClass + "'", e);
            }

            return current;

        }

        // else return default implementation
        return new SecurityCurrent();
    }

    /**
     * Return the unique instance of this object.
     * @return SecurityCurrent return the current
     */
    public static EZBSecurityCurrent getCurrent() {
        return unique;
    }

    /**
     * Associates the given security context to the current thread.
     * @param securityContext Security context to associate to the current
     *        thread.
     */
    public void setSecurityContext(final EZBSecurityContext securityContext) {
        threadLocal.set(securityContext);
    }

    /**
     * Associates the given security context to all threads (JVM).
     * @param securityContext Security context to associate to the JVM
     */
    public void setGlobalSecurityContext(final EZBSecurityContext securityContext) {
        globalContext = securityContext;
    }

    /**
     * Gets the current context.
     * @return SecurityContext return the Security context associated to the
     *         current thread or the JVM
     */
    public EZBSecurityContext getSecurityContext() {
        if (globalContext != null) {
            return globalContext;
        }
        if (threadLocal.get() != null) {
            return threadLocal.get();
        }

        // else, never null context.
        return DEFAULT_CTX;
    }

}
