/**
 * EasyBeans
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id:$
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.deployer;

import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.List;

import org.objectweb.easybeans.api.EZBArchiveException;
import org.objectweb.easybeans.api.EZBContainer;
import org.objectweb.easybeans.api.EZBContainerException;
import org.objectweb.easybeans.deployer.api.CARDeployable;
import org.objectweb.easybeans.deployer.api.EARDeployable;
import org.objectweb.easybeans.deployer.api.EJB3Deployable;
import org.objectweb.easybeans.deployer.api.EJBDeployable;
import org.objectweb.easybeans.deployer.api.EZBDeployable;
import org.objectweb.easybeans.deployer.api.EZBDeployerException;
import org.objectweb.easybeans.deployer.api.LibDeployable;
import org.objectweb.easybeans.deployer.api.WARDeployable;
import org.objectweb.easybeans.loader.EasyBeansClassLoader;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;
import org.objectweb.easybeans.persistence.PersistenceUnitManager;
import org.objectweb.easybeans.persistence.xml.PersistenceXmlFileAnalyzer;
import org.objectweb.easybeans.persistence.xml.PersistenceXmlFileAnalyzerException;

/**
 * This class manage the deployment of EAR for web container. It extracts EJB3
 * and send them to EasyBeans while the War are given to the web container.
 * @author Florent Benoit
 */
public abstract class AbsWebContainerDeployer extends AbsDeployer {

    /**
     * Logger.
     */
    private static JLog logger = JLogFactory.getLog(AbsWebContainerDeployer.class);

    /**
     * Deploy the WAR files present in the given EAR.
     * @param earDeployable the EAR containing the WARs
     * @param earURL the EAR URL
     * @param earClassLoader the EAR classloader
     * @param parentClassLoader the parent classloader (EJB) to use
     * @throws EZBDeployerException if the wars are not deployed.
     */
    protected abstract void deployWARs(final EARDeployable earDeployable, final URL earURL, final ClassLoader earClassLoader,
            final ClassLoader parentClassLoader) throws EZBDeployerException;

    /**
     * Deploy an EAR (called by the deploy method).
     * @param earDeployable a given EAR deployable
     * @throws EZBDeployerException if the deployment is not done.
     */
    protected void deployEAR(final EARDeployable earDeployable) throws EZBDeployerException {

        logger.info("Deploying {0}", earDeployable);

        // Get URL of this EAR
        URL earURL = null;
        try {
            earURL = earDeployable.getArchive().getURL();
        } catch (EZBArchiveException e) {
            throw new EZBDeployerException("Cannot get the URL for the deployable '" + earDeployable + "'.", e);
        }

        // Create a Root classloader for this EAR
        // Empty classloader
        ClassLoader earClassLoader = new URLClassLoader(new URL[0], Thread.currentThread().getContextClassLoader());

        // Get EJBs of this EAR
        List<EJB3Deployable> ejb3s = earDeployable.getEJB3Deployables();

        // Get the URLs of EJB, WEB and Clients
        List<URL> urlsEJB = new ArrayList<URL>();
        for (EJB3Deployable ejb : ejb3s) {
            try {
                urlsEJB.add(ejb.getArchive().getURL());
            } catch (EZBArchiveException e) {
                throw new EZBDeployerException("Cannot get the URL for the archive '" + ejb.getArchive() + "'", e);
            }
        }
        List<URL> urlsWAR = new ArrayList<URL>();
        for (WARDeployable war : earDeployable.getWARDeployables()) {
            try {
                urlsWAR.add(war.getArchive().getURL());
            } catch (EZBArchiveException e) {
                throw new EZBDeployerException("Cannot get the URL for the archive '" + war.getArchive() + "'", e);
            }
        }
        List<URL> urlsClient = new ArrayList<URL>();
        for (CARDeployable car : earDeployable.getCARDeployables()) {
            try {
                urlsClient.add(car.getArchive().getURL());
            } catch (EZBArchiveException e) {
                throw new EZBDeployerException("Cannot get the URL for the archive '" + car.getArchive() + "'", e);
            }
        }

        // Get libraries of this EAR
        List<LibDeployable> libs = earDeployable.getLibDeployables();

        // Create array of URLs with EJBs + Libraries
        List<URL> urls = new ArrayList<URL>();
        for (EJBDeployable ejb : ejb3s) {
            try {
                urls.add(ejb.getArchive().getURL());
            } catch (EZBArchiveException e) {
                throw new EZBDeployerException("Cannot get the URL for the Archive '" + ejb.getArchive() + "'.", e);
            }
        }
        for (LibDeployable lib : libs) {
            try {
                urls.add(lib.getArchive().getURL());
            } catch (EZBArchiveException e) {
                throw new EZBDeployerException("Cannot get the URL for the Archive '" + lib.getArchive() + "'.", e);

            }
        }

        // Create classloader with these URLs
        URL[] arrayURLs = urls.toArray(new URL[urls.size()]);
        ClassLoader ejbClassLoader = new EasyBeansClassLoader(arrayURLs, Thread.currentThread().getContextClassLoader());

        // Analyze libraries to detect persistence archive (only once for now
        // and for all libraries)
        PersistenceUnitManager persistenceUnitManager = null;
        for (LibDeployable lib : libs) {
            PersistenceUnitManager builtPersistenceUnitManager = null;
            try {
                builtPersistenceUnitManager = PersistenceXmlFileAnalyzer.analyzePersistenceXmlFile(lib.getArchive(),
                        ejbClassLoader);
            } catch (PersistenceXmlFileAnalyzerException e) {
                throw new IllegalStateException("Failure when analyzing the persistence.xml file", e);
            }
            if (persistenceUnitManager != null) {
                throw new IllegalStateException("There is an existing persistence.xml in another library of this EAR."
                        + " This is not supported.");
            }
            persistenceUnitManager = builtPersistenceUnitManager;
        }

        // Create containers for each EJB deployable
        List<EZBContainer> containers = new ArrayList<EZBContainer>();
        for (EJBDeployable ejb : ejb3s) {
            containers.add(getEmbedded().createContainer(ejb.getArchive()));
        }

        // Configure containers
        for (EZBContainer container : containers) {
            // Set the classloader that needs to be used
            container.setClassLoader(ejbClassLoader);
            // Add persistence context found
            container.setPersistenceUnitManager(persistenceUnitManager);
        }

        // Start containers
        for (EZBContainer container : containers) {
            try {
                container.start();
            } catch (EZBContainerException e) {
                logger.error("Cannot start container {0}", container.getName(), e);
            }
        }

        // Deploy Web App
        deployWARs(earDeployable, earURL, earClassLoader, ejbClassLoader);

    }

    /**
     * Undeploy an EAR (called by the undeploy method).
     * @param tmpEARDeployable a given EAR deployable
     * @throws EZBDeployerException if the undeployment is not done.
     */
    protected void undeployEAR(final EARDeployable tmpEARDeployable) throws EZBDeployerException {
        logger.info("Undeploying {0}", tmpEARDeployable);

        // From which deployable get the containers deployed
        EARDeployable earDeployable = tmpEARDeployable;

        // Check if this archive has been unpacked ?
        EARDeployable unpackedDeployable = earDeployable.getUnpackedDeployable();
        if (unpackedDeployable != null) {
            earDeployable = unpackedDeployable;
        }

        // Need to undeploy Wars from the EAR
        List<WARDeployable> wars = earDeployable.getWARDeployables();
        if (wars != null) {
            for (WARDeployable war : wars) {
                // undeploy the given war
                try {
                    undeployWAR(war);
                } catch (EZBDeployerException e) {
                    logger.error("Cannot undeploy the WAR deployable ''{0}''", war, e);
                }
            }
        }

        // Undeploy EJB3s
        undeployEJB3FromEAR(earDeployable);

        logger.info("''{0}'' EAR Deployable is now undeployed", tmpEARDeployable);
    }

    /**
     * Undeploy the given deployable. It can be an EJB jar, EAR, WAR, etc.
     * @param deployable a given deployable to undeploy
     * @throws EZBDeployerException if the undeploy operation fails.
     */
    public void undeploy(final EZBDeployable deployable) throws EZBDeployerException {
        if (deployable instanceof EARDeployable) {
            undeployEAR((EARDeployable) deployable);
        } else {
            throw new UnsupportedOperationException("Undeploy only .ear files");
        }
    }

    /**
     * Undeploy an given WAR (called by the undeploy method).
     * @param warDeployable a given WAR deployable
     * @throws EZBDeployerException if the undeployment is not done.
     */
    protected abstract void undeployWAR(final WARDeployable warDeployable) throws EZBDeployerException;
}
