/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: MetadataMerge.java 1252 2006-12-14 10:32:05Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.deployment.xml.helper;

import static org.objectweb.easybeans.deployment.xml.struct.common.LifeCycleCallback.POST_ACTIVATE;
import static org.objectweb.easybeans.deployment.xml.struct.common.LifeCycleCallback.POST_CONSTRUCT;
import static org.objectweb.easybeans.deployment.xml.struct.common.LifeCycleCallback.PRE_DESTROY;
import static org.objectweb.easybeans.deployment.xml.struct.common.LifeCycleCallback.PRE_PASSIVATE;

import java.util.ArrayList;
import java.util.List;

import javax.ejb.Remove;
import javax.ejb.TransactionManagementType;

import org.objectweb.asm.Type;
import org.objectweb.easybeans.deployment.annotations.ClassType;
import org.objectweb.easybeans.deployment.annotations.impl.JAnnotationResource;
import org.objectweb.easybeans.deployment.annotations.impl.JApplicationException;
import org.objectweb.easybeans.deployment.annotations.impl.JCommonBean;
import org.objectweb.easybeans.deployment.annotations.impl.JEjbEJB;
import org.objectweb.easybeans.deployment.annotations.impl.JInterceptors;
import org.objectweb.easybeans.deployment.annotations.impl.JLocal;
import org.objectweb.easybeans.deployment.annotations.impl.JRemote;
import org.objectweb.easybeans.deployment.annotations.impl.JRemove;
import org.objectweb.easybeans.deployment.annotations.metadata.ClassAnnotationMetadata;
import org.objectweb.easybeans.deployment.annotations.metadata.EjbJarAnnotationMetadata;
import org.objectweb.easybeans.deployment.annotations.metadata.FieldAnnotationMetadata;
import org.objectweb.easybeans.deployment.annotations.metadata.MethodAnnotationMetadata;
import org.objectweb.easybeans.deployment.xml.struct.AbsBean;
import org.objectweb.easybeans.deployment.xml.struct.AssemblyDescriptor;
import org.objectweb.easybeans.deployment.xml.struct.EJB3;
import org.objectweb.easybeans.deployment.xml.struct.EnterpriseBeans;
import org.objectweb.easybeans.deployment.xml.struct.Interceptor;
import org.objectweb.easybeans.deployment.xml.struct.InterceptorBinding;
import org.objectweb.easybeans.deployment.xml.struct.Interceptors;
import org.objectweb.easybeans.deployment.xml.struct.RemoveMethod;
import org.objectweb.easybeans.deployment.xml.struct.Session;
import org.objectweb.easybeans.deployment.xml.struct.common.AbsEJBRef;
import org.objectweb.easybeans.deployment.xml.struct.common.AbsEnvironment;
import org.objectweb.easybeans.deployment.xml.struct.common.AroundInvoke;
import org.objectweb.easybeans.deployment.xml.struct.common.EJBLocalRef;
import org.objectweb.easybeans.deployment.xml.struct.common.EJBRef;
import org.objectweb.easybeans.deployment.xml.struct.common.InjectionTarget;
import org.objectweb.easybeans.deployment.xml.struct.common.LifeCycleCallback;
import org.objectweb.easybeans.deployment.xml.struct.common.MethodDD;
import org.objectweb.easybeans.deployment.xml.struct.common.ResourceEnvRef;
import org.objectweb.easybeans.deployment.xml.struct.common.ResourceRef;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;

/**
 * This class merge the information of the struct representing EJB DD into the
 * metadata.
 * @author Florent Benoit
 */
public final class MetadataMerge {

    /**
     * Wildcard attribute.
     */
    private static final String WILDCARD = "*";

    /**
     * Defines java.lang.Object class.
     */
    public static final String JAVA_LANG_OBJECT = "java/lang/Object";

    /**
     * Metadata for an ejb-jar.
     */
    private EjbJarAnnotationMetadata ejbMetadata = null;

    /**
     * Logger.
     */
    private static JLog logger = JLogFactory.getLog(MetadataMerge.class);

    /**
     * Helper class, no public constructor.
     * @param ejbMetadata the metadata corresponding to an EJB-JAR file.
     */
    private MetadataMerge(final EjbJarAnnotationMetadata ejbMetadata) {
        this.ejbMetadata = ejbMetadata;
    }

    /**
     * Takes struct of metadata and adds/set information on class/methods
     * metadata.
     * @param ejbMetadata the metadata corresponding to an EJB-JAR file.
     */
    public static void merge(final EjbJarAnnotationMetadata ejbMetadata) {
        new MetadataMerge(ejbMetadata).resolve();
    }

    /**
     * Do all merging operations.
     */
    private void resolve() {

        EJB3 ejb3 = ejbMetadata.getEjb3();
        if (ejb3 != null) {
            // Get enterprise beans
            EnterpriseBeans enterpriseBeans = ejb3.getEnterpriseBeans();
            if (enterpriseBeans != null) {

                // get session beans
                List<Session> sessionList = enterpriseBeans.getSessionList();
                for (Session session : sessionList) {
                    // get ejb-name
                    String ejbName = session.getEjbName();
                    String ejbClass = session.getEjbClass();
                    ClassAnnotationMetadata classAnnotationMetadata = findClassForEjb(ejbName, ejbClass);

                    // class is found, then apply all customization.
                    applySessionBean(session, classAnnotationMetadata);
                }
            }

            // Interceptors
            Interceptors interceptors = ejb3.getInterceptors();
            if (interceptors != null) {
                applyInterceptors(interceptors);
            }

            // Assembly descriptor
            AssemblyDescriptor assemblyDescriptor = ejb3.getAssemblyDescriptor();
            if (assemblyDescriptor != null) {

                // merge application-exception
                mergeApplicationException(assemblyDescriptor);

                // merge interceptor-binding
                mergeInterceptorBinding(assemblyDescriptor);

            }
        }
    }

    /**
     * Apply rules for the interceptors defined in the XML DD.
     * @param interceptors the interceptors struct representing
     *        &lt;interceptors&gt; element
     */
    private void applyInterceptors(final Interceptors interceptors) {
        // Get interceptors
        List<Interceptor> interceptorList = interceptors.getInterceptorList();
        if (interceptorList != null) {
            for (Interceptor interceptor : interceptorList) {

                // Check that there is an interceptor class defined
                String itCptClass = interceptor.getInterceptorClass();
                if (itCptClass == null) {
                    throw new IllegalStateException("Invalid interceptor with name '" + interceptor + "': No interceptor-class");
                }

                // Class Metadata
                ClassAnnotationMetadata itcpMetadata = ejbMetadata.getClassAnnotationMetadata(encode(itCptClass));
                if (itcpMetadata == null) {
                    throw new IllegalStateException("No metadata found for class '" + itCptClass + "'.");
                }

                // Will use metadata of the interceptor class
                applyJndiEnvironmentRefsGroup(interceptor, itcpMetadata);

            }
        }
    }

    /**
     * Apply rules for a session bean.
     * @param sessionbean the struct (XML DD)
     * @param classAnnotationMetadata (Annotation metadata)
     */
    private void applySessionBean(final Session sessionbean, final ClassAnnotationMetadata classAnnotationMetadata) {
        applyCommonBean(sessionbean, classAnnotationMetadata);

        // set session-type
        String sessionType = sessionbean.getSessionType();
        if (sessionType == null && !classAnnotationMetadata.isSession()) {
            throw new IllegalStateException("Missing session-type for bean '" + sessionbean + "' and no annotation in '"
                    + classAnnotationMetadata + "'.");
        }
        // set the type of bean
        if ("Stateless".equals(sessionType)) {
            classAnnotationMetadata.setClassType(ClassType.STATELESS);
        } else if ("Stateful".equals(sessionType)) {
            classAnnotationMetadata.setClassType(ClassType.STATEFUL);
        }

        // Transaction type
        String transactionType = sessionbean.getTransactionType();
        if (transactionType != null) {
            if ("Bean".equals(transactionType)) {
                classAnnotationMetadata.setTransactionManagementType(TransactionManagementType.BEAN);
            } else if ("Container".equals(transactionType)) {
                classAnnotationMetadata.setTransactionManagementType(TransactionManagementType.CONTAINER);
            } else {
                throw new IllegalStateException("Invalid transaction type '" + transactionType + "' found");
            }
        }

        // Remote Home
        String remoteHome = sessionbean.getRemoteHome();
        if (remoteHome != null) {
            classAnnotationMetadata.setRemoteHome(remoteHome);
        }

        // Local Home
        String localHome = sessionbean.getLocalHome();
        if (localHome != null) {
            classAnnotationMetadata.setLocalHome(localHome);
        }

        // local interface
        if (sessionbean.getBusinessLocalList().size() > 0) {
            JLocal jLocal = classAnnotationMetadata.getLocalInterfaces();
            if (jLocal == null) {
                jLocal = new JLocal();
                classAnnotationMetadata.setLocalInterfaces(jLocal);
            }
            for (String itf : sessionbean.getBusinessLocalList()) {
                String encodedItf = encode(itf);
                if (!jLocal.contains(encodedItf)) {
                    jLocal.addInterface(encodedItf);
                }
            }
        }

        // Mapped name
        if (sessionbean.getMappedName() != null) {
            JCommonBean jCommonBean = classAnnotationMetadata.getJCommonBean();
            if (jCommonBean == null) {
                throw new IllegalStateException("No JCommonBean attribute on class '" + classAnnotationMetadata.getClassName()
                        + "' while there is a mapped name.");
            }
            jCommonBean.setMappedName(sessionbean.getMappedName());
        }

        // remote interface
        if (sessionbean.getBusinessRemoteList().size() > 0) {
            JRemote jRemote = classAnnotationMetadata.getRemoteInterfaces();
            if (jRemote == null) {
                jRemote = new JRemote();
                classAnnotationMetadata.setRemoteInterfaces(jRemote);
            }
            for (String itf : sessionbean.getBusinessRemoteList()) {
                String encodedItf = encode(itf);
                if (!jRemote.contains(encodedItf)) {
                    jRemote.addInterface(encodedItf);
                }
            }
        }

        // Remove methods (for stateful bean)
        if (sessionbean.getRemoveMethodList().size() > 0) {
            if (!classAnnotationMetadata.isStateful()) {
                throw new IllegalStateException(
                        "Cannot apply XML DD remove methods on a bean which is not a stateful bean. Bean class is '"
                                + classAnnotationMetadata.getClassName() + "'. Remove Methods are '"
                                + sessionbean.getRemoveMethodList() + "'.");
            }

            for (RemoveMethod removeMethod : sessionbean.getRemoveMethodList()) {
                MethodDD method = removeMethod.getMethod();

                // method case
                List<MethodAnnotationMetadata> methods = getMethodsForGivenMethodDD(method, classAnnotationMetadata);
                // No method found (warning)
                if (methods.size() == 0) {
                    logger.warn("No method found for the remove-method ''{0}'' on the class ''{1}''", method,
                            classAnnotationMetadata.getClassName());
                } else if (methods.size() > 1) {
                    logger.warn("Found more than 1 method for the remove-method ''{0}'' on the class ''{1}''."
                            + "All these methods will be set as remove methods.", method, classAnnotationMetadata.getClassName());
                }

                // Set as a remove method
                for (MethodAnnotationMetadata methodAnnotationMetadata : methods) {

                    // Get existing value
                    Remove previousRemove = methodAnnotationMetadata.getJRemove();

                    // Value of the XML (for retain)
                    Boolean retainXML = removeMethod.getRetainIfException();

                    // If value has been defined in the XML, use this value,
                    // else use the value of the annotation, else the default
                    // value (false)
                    boolean retain = false;
                    if (retainXML != null) {
                        retain = retainXML.booleanValue();
                    } else if (previousRemove != null) {
                        // got value of the annotation
                        retain = previousRemove.retainIfException();
                    }

                    // Set the new value
                    JRemove jRemove = new JRemove(retain);
                    methodAnnotationMetadata.setRemove(jRemove);
                }
            }
        }
    }

    /**
     * Apply rules for a common bean.
     * @param bean the struct (XML DD)
     * @param classAnnotationMetadata (Annotation metadata)
     */
    private void applyCommonBean(final AbsBean bean, final ClassAnnotationMetadata classAnnotationMetadata) {
        // no Common bean, create one.
        if (classAnnotationMetadata.getJCommonBean() == null) {
            classAnnotationMetadata.setJCommonBean(new JCommonBean());
        }
        // Sets the ejb-name
        JCommonBean jCommonBean = classAnnotationMetadata.getJCommonBean();
        boolean isDefaultName = jCommonBean.isDefaultName();
        if (isDefaultName) {
            jCommonBean.setName(bean.getEjbName());
            jCommonBean.setDefaultName(false);
        } else {
            // Bean already registered with this name, add an alias
            jCommonBean.addAlias(bean.getEjbName());
        }

        applyJndiEnvironmentRefsGroup(bean, classAnnotationMetadata);

    }

    /**
     * Apply rules for a an environment.
     * @param envGroup the struct (XML DD)
     * @param classAnnotationMetadata (Annotation metadata)
     */
    private void applyJndiEnvironmentRefsGroup(final AbsEnvironment envGroup,
            final ClassAnnotationMetadata classAnnotationMetadata) {

        // post-construct / pre-destroy
        applyLifeCycle(envGroup.getPostConstructCallbackList(), classAnnotationMetadata, POST_CONSTRUCT);
        applyLifeCycle(envGroup.getPreDestroyCallbackList(), classAnnotationMetadata, PRE_DESTROY);
        applyLifeCycle(envGroup.getPrePassivateCallbackList(), classAnnotationMetadata, PRE_PASSIVATE);
        applyLifeCycle(envGroup.getPostActivateCallbackList(), classAnnotationMetadata, POST_ACTIVATE);
        applyAroundInvoke(envGroup.getAroundInvokeList(), classAnnotationMetadata, AroundInvoke.NAME);

        // ejb(local) ref
        applyEjbRef(envGroup.getEJBRefList(), classAnnotationMetadata);
        applyEjbLocalRef(envGroup.getEJBLocalRefList(), classAnnotationMetadata);

        // resource(env) ref
        applyResourceRef(envGroup.getResourceRefList(), classAnnotationMetadata);
        applyResourceEnvRef(envGroup.getResourceEnvRefList(), classAnnotationMetadata);
    }

    /**
     * Apply rules for a common bean (resource-ref).
     * @param resourceRefList the list of resource-ref.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     */
    private void applyResourceRef(final List<ResourceRef> resourceRefList,
            final ClassAnnotationMetadata classAnnotationMetadata) {
        if (resourceRefList == null) {
            return;
        }

        // For each resource-ref, build a &#64;{@link javax.annotation.Resource}
        for (ResourceRef resourceRef : resourceRefList) {
            JAnnotationResource jResource = new JAnnotationResource();
            // set name
            jResource.setName(resourceRef.getResRefName());

            // set type
            jResource.setType(resourceRef.getResRefType());

            // Apply the resource on the class
            applyJResource(jResource, classAnnotationMetadata, resourceRef.getInjectionTargetList());
        }
    }

    /**
     * Apply rules for a common bean (resource-env-ref).
     * @param resourceEnvRefList the list of resource-env-ref.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     */
    private void applyResourceEnvRef(final List<ResourceEnvRef> resourceEnvRefList,
            final ClassAnnotationMetadata classAnnotationMetadata) {
        if (resourceEnvRefList == null) {
            return;
        }

        // For each resource-ref, build a &#64;{@link javax.annotation.Resource}
        for (ResourceEnvRef resourceEnvRef : resourceEnvRefList) {
            JAnnotationResource jResource = new JAnnotationResource();
            // set name
            jResource.setName(resourceEnvRef.getResourceEnvRefName());

            // set type
            jResource.setType(resourceEnvRef.getResourceEnvRefType());

            // Apply the resource on the class
            applyJResource(jResource, classAnnotationMetadata, resourceEnvRef.getInjectionTargetList());
        }
    }

    /**
     * Apply rules for a common bean (ejb-ref).
     * @param ejbRefList the list of ejb-ref.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     */
    private void applyEjbRef(final List<EJBRef> ejbRefList,
            final ClassAnnotationMetadata classAnnotationMetadata) {
        if (ejbRefList == null) {
            return;
        }

        // For each ejb-ref, build a &#64;{@link javax.ejb.EJB}
        for (EJBRef ejbRef : ejbRefList) {
            JEjbEJB jEJB = buildEJB(ejbRef);
            // set interface
            if (ejbRef.getHome() != null) {
                jEJB.setBeanInterface(ejbRef.getHome());
            } else if (ejbRef.getRemote() != null) {
                jEJB.setBeanInterface(ejbRef.getRemote());
            }
            applyJEJB(jEJB, classAnnotationMetadata, ejbRef.getInjectionTargetList());
        }
    }

    /**
     * Apply rules for a common bean (ejb-local-ref).
     * @param ejbLocalRefList the list of ejb-local-ref.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     */
    private void applyEjbLocalRef(final List<EJBLocalRef> ejbLocalRefList,
            final ClassAnnotationMetadata classAnnotationMetadata) {
        if (ejbLocalRefList == null) {
            return;
        }

        // For each ejb-local-ref, build a &#64;{@link javax.ejb.EJB}
        for (EJBLocalRef ejbLocalRef : ejbLocalRefList) {
            JEjbEJB jEJB = buildEJB(ejbLocalRef);
            // set interface
            if (ejbLocalRef.getLocalHome() != null) {
                jEJB.setBeanInterface(ejbLocalRef.getLocalHome());
            } else if (ejbLocalRef.getLocal() != null) {
                jEJB.setBeanInterface(ejbLocalRef.getLocal());
            }
            applyJEJB(jEJB, classAnnotationMetadata, ejbLocalRef.getInjectionTargetList());
        }
    }


    /**
     * Adds the given &#64;{@link javax.annotation.Resource} object to the class and apply it also on injection target elements.
     * @param jResource an &#64;{@link javax.annotation.Resource} impl
     * @param classAnnotationMetadata the bean's class
     * @param injectionTargetList the list of injection fields to set
     */
    private static void applyJResource(final JAnnotationResource jResource, final ClassAnnotationMetadata classAnnotationMetadata,
            final List<InjectionTarget> injectionTargetList) {

        // Add field on the class
        JAnnotationResource existingJResource = classAnnotationMetadata.getJAnnotationResource();

        // Needs to use a list instead.
        List<JAnnotationResource> jResources = classAnnotationMetadata.getJAnnotationResources();
        if (jResources == null) {
            jResources = new ArrayList<JAnnotationResource>();
            classAnnotationMetadata.setJAnnotationResources(jResources);
        }

        // put existing element in the list
        if (existingJResource != null) {
            jResources.add(existingJResource);
            // remove single element
            classAnnotationMetadata.setJAnnotationResource(null);
        }

        // Add our new element
        jResources.add(jResource);

        // Now, add elements on the injection fields
        if (injectionTargetList != null) {
            for (InjectionTarget injectionTarget : injectionTargetList) {

                // Add the JAnnotationResource on it
                getField(classAnnotationMetadata, injectionTarget).setJAnnotationResource(jResource);
            }
        }


    }

    /**
     * Gets the field metadata for a given injection target.
     * @param classAnnotationMetadata the class used to search the target
     *        element
     * @param injectionTarget the given target info element
     * @return a field if element is found
     */
    private static FieldAnnotationMetadata getField(final ClassAnnotationMetadata classAnnotationMetadata,
            final InjectionTarget injectionTarget) {
        // get data from the injection target element
        String clName = injectionTarget.getClassname();
        if (clName == null) {
            clName = classAnnotationMetadata.getClassName();
        } else {
            clName = clName.replace(".", "/");
        }

        String fieldName = injectionTarget.getTargetName();

        // Now, search the element in the given class
        ClassAnnotationMetadata foundClassMetadata = classAnnotationMetadata.getEjbJarAnnotationMetadata()
        .getClassAnnotationMetadata(clName);
        if (foundClassMetadata == null) {
            throw new IllegalArgumentException("Cannot find the class metadata named '" + clName + "'.");
        }

        // Got the class metadata, now search the field
        List<FieldAnnotationMetadata> fields = foundClassMetadata.searchFieldAnnotationMetadata(fieldName);

        // Should be only one
        if (fields.size() == 0) {
            throw new IllegalArgumentException("Cannot find field with name '" + fieldName + "' in the class '" + clName
                    + "'.");
        } else if (fields.size() > 1) {
            throw new IllegalArgumentException("Too many fields with name '" + fieldName + "' found in the class '"
                    + clName + "'.");
        }

        return fields.get(0);
    }


    /**
     * Adds the given &#64;{@link javax.ejb.EJB} object to the class and apply it also on injection target elements.
     * @param jEJB an &#64;{@link javax.ejb.EJB} impl
     * @param classAnnotationMetadata the bean's class
     * @param injectionTargetList the list of injection fields to set
     */
    private static void applyJEJB(final JEjbEJB jEJB, final ClassAnnotationMetadata classAnnotationMetadata,
            final List<InjectionTarget> injectionTargetList) {

        // Add field on the class
        JEjbEJB existingJEjbEJB = classAnnotationMetadata.getJEjbEJB();

        // Needs to use a list instead.
        List<JEjbEJB> jEjbEJBs = classAnnotationMetadata.getJEjbEJBs();
        if (jEjbEJBs == null) {
            jEjbEJBs = new ArrayList<JEjbEJB>();
            classAnnotationMetadata.setJEjbEJBs(jEjbEJBs);
        }

        // put existing element in the list
        if (existingJEjbEJB != null) {
            jEjbEJBs.add(existingJEjbEJB);
            // remove single element
            classAnnotationMetadata.setJEjbEJB(null);
        }

        // Add our new element
        jEjbEJBs.add(jEJB);

        // Now, add elements on the injection fields
        if (injectionTargetList != null) {
            for (InjectionTarget injectionTarget : injectionTargetList) {

                // Add the JEjbEJB on it
                getField(classAnnotationMetadata, injectionTarget).setJEjbEJB(jEJB);
            }
        }

    }



    /**
     * Build an &#64;{@link javax.ejb.EJB} impl for the given ejb-ref/ejb-local-ref.
     * @param ref an instance of a ejb-ref/ejb-local-ref.
     * @return a implementation of &#64;{@link javax.ejb.EJB} interface.
     */
    private static JEjbEJB buildEJB(final AbsEJBRef ref) {
        JEjbEJB jEJB = new JEjbEJB();
        // Sets the name
        if (ref.getEjbRefName() != null) {
            jEJB.setName(ref.getEjbRefName());
        }

        // sets the mapped-name
        if (ref.getMappedName() != null) {
            jEJB.setMappedName(ref.getMappedName());
        }

        // sets the bean-name
        if (ref.getEjbLink() != null) {
            jEJB.setBeanName(ref.getEjbLink());
        }

        return jEJB;
    }


    /**
     * Apply rules for a common bean (method callbacks).
     * @param aroundInvokeList the list of method/aroundInvoke callbacks.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     * @param type the type of lifecycle.
     */
    private void applyAroundInvoke(final List<AroundInvoke> aroundInvokeList,
            final ClassAnnotationMetadata classAnnotationMetadata, final String type) {
        for (AroundInvoke aroundInvoke : aroundInvokeList) {
            String interceptorClassName = null;

            // if no class, takes the bean's class
            if (aroundInvoke.getClassName() == null) {
                interceptorClassName = classAnnotationMetadata.getClassName();
            } else {
                interceptorClassName = encode(aroundInvoke.getClassName());
            }

            // Now, search the Method on the class or its super-class if not
            // found
            String searchMethodName = aroundInvoke.getMethodName();
            if (searchMethodName == null) {
                throw new IllegalStateException("Invalid null method name for the callback '" + interceptorClassName + "'.");
            }

            // Apply interceptor
            applyInterceptor(type, interceptorClassName, searchMethodName, classAnnotationMetadata);

        }
    }


    /**
     * Apply rules for a common bean (lifecycle callbacks).
     * @param lifecycleList the list of lifecycle callbacks.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     * @param type the type of lifecycle.
     */
    private void applyLifeCycle(final List<LifeCycleCallback> lifecycleList,
            final ClassAnnotationMetadata classAnnotationMetadata, final String type) {
        for (LifeCycleCallback lifecycleCallback : lifecycleList) {

            String interceptorClassName = null;

            // if no class, takes the bean's class
            if (lifecycleCallback.getLifecycleCallbackClass() == null) {
                interceptorClassName = classAnnotationMetadata.getClassName();
            } else {
                interceptorClassName = encode(lifecycleCallback.getLifecycleCallbackClass());
            }

            // Now, search the Method on the class or its super-class if not
            // found
            String searchMethodName = lifecycleCallback.getLifecycleCallbackMethod();
            if (searchMethodName == null) {
                throw new IllegalStateException("Invalid null method name for the callback '" + interceptorClassName + "'.");
            }

            // Apply interceptor
            applyInterceptor(type, interceptorClassName, searchMethodName, classAnnotationMetadata);

        }
    }

    /**
     * Adds a link on the given class to the interceptor class.
     * @param classAnnotationMetadata the class that will use/import the
     *        interceptor
     * @param interceptorClassName the interceptor class.
     */
    private void addInterceptorForClass(final ClassAnnotationMetadata classAnnotationMetadata,
            final String interceptorClassName) {
        // Get interceptors for the given class (the bean class)
        JInterceptors interceptors = classAnnotationMetadata.getAnnotationInterceptors();
        if (interceptors == null) {
            // build new list if empty
            interceptors = new JInterceptors();
            classAnnotationMetadata.setAnnotationsInterceptors(interceptors);
        }

        // class already present ?
        if (!interceptors.contains(interceptorClassName)) {
            interceptors.addClass(interceptorClassName);
        }
    }

    /**
     * Finds a method in the interceptor class and set the correct type value.
     * @param type type of interceptor
     * @param interceptorClassName the name of the interceptor class
     * @param methodName the name of the method in the interceptor class
     * @param classAnnotationMetadata the class that is using the interceptor (may be null)
     */
    private void applyInterceptor(final String type, final String interceptorClassName, final String methodName,
            final ClassAnnotationMetadata classAnnotationMetadata) {

        // if the bean is using this interceptor and that this is not the bean class, add the link.
        if (classAnnotationMetadata != null  && !classAnnotationMetadata.getClassName().equals(interceptorClassName)) {
            addInterceptorForClass(classAnnotationMetadata, interceptorClassName);
        }

        MethodAnnotationMetadata foundMethod = null;
        ClassAnnotationMetadata interceptorMetadata = ejbMetadata.getClassAnnotationMetadata(interceptorClassName);


        boolean loop = true;
        // While method is not found and that class has super class, try to
        // search
        while (foundMethod == null && loop) {
            List<MethodAnnotationMetadata> listMethods = interceptorMetadata.searchMethodAnnotationMetadata(methodName);

            // Method present ?
            if (listMethods.size() == 1) {
                // first item
                foundMethod = listMethods.get(0);
                continue;
            } else if (listMethods.size() > 1) {
                throw new IllegalStateException("Method with name '" + methodName + "' was found more than once on the class '"
                        + interceptorMetadata.getClassName() + "'.");
            }

            // else continue
            String superClassName = interceptorMetadata.getSuperName();
            if (!superClassName.equals(JAVA_LANG_OBJECT)) {
                interceptorMetadata = ejbMetadata.getClassAnnotationMetadata(superClassName);
            } else {
                // break loop
                loop = false;
            }
        }

        if (foundMethod == null) {
            throw new IllegalStateException("The method named '" + methodName + "' was not found in the class '"
                    + interceptorMetadata.getClassName() + "' and all its super-classes.");
        }

        // call setter to true and only if it is not already present.
        if (POST_CONSTRUCT.equals(type)) {
            if (!foundMethod.isPostConstruct()) {
                foundMethod.setPostConstruct(true);
                interceptorMetadata.addPostConstructMethodMetadata(foundMethod);
            }
        } else if (PRE_DESTROY.equals(type)) {
            if (!foundMethod.isPreDestroy()) {
                foundMethod.setPreDestroy(true);
                interceptorMetadata.addPreDestroyMethodMetadata(foundMethod);
            }
        } else if (PRE_PASSIVATE.equals(type)) {
            if (!foundMethod.isPrePassivate()) {
                foundMethod.setPrePassivate(true);
                interceptorMetadata.addPrePassivateMethodMetadata(foundMethod);
            }
        } else if (POST_ACTIVATE.equals(type)) {
            if (!foundMethod.isPostActivate()) {
                foundMethod.setPostActivate(true);
                interceptorMetadata.addPostActivateMethodMetadata(foundMethod);
            }
        } else if (AroundInvoke.NAME.equals(type)) {
            if (!foundMethod.isAroundInvoke()) {
                foundMethod.setAroundInvoke(true);
                interceptorMetadata.addAroundInvokeMethodMetadata(foundMethod);
            }
        } else {
            throw new IllegalStateException("No case for type '" + type + "'..");
        }
    }

    /**
     * Find/Gets a classAnnotationMetadata for a given ejb name and/or a given
     * ejbclass name.
     * @param ejbName the name of the EJB.
     * @param ejbClass the class of the EJB.
     * @return a classAnnotationMetadata if found, else exception
     */
    private ClassAnnotationMetadata findClassForEjb(final String ejbName, final String ejbClass) {
        ClassAnnotationMetadata classAnnotationMetadata = null;

        // First, try with ejbClass if not null
        if (ejbClass != null) {
            classAnnotationMetadata = ejbMetadata.getClassAnnotationMetadata(encode(ejbClass));
            // found ? return it
            if (classAnnotationMetadata != null) {
                return classAnnotationMetadata;
            }
        }

        // else, needs to find a class with the specified name
        for (ClassAnnotationMetadata tmpMetaData : ejbMetadata.getClassAnnotationMetadataCollection()) {
            if (tmpMetaData.isBean()) {
                // check name
                JCommonBean bean = tmpMetaData.getJCommonBean();
                if (bean != null) {
                    // found !
                    if (ejbName != null && ejbName.equals(bean.getName())) {
                        return tmpMetaData;
                    }
                }
            }
        }
        throw new IllegalStateException("No class with ejb-name '" + ejbName + "' or ejb-class '" + ejbClass
                + "' was found in the EJB-JAR file '" + ejbMetadata);
    }

    /**
     * Set/override the current application-exception.
     * @param assemblyDescriptor the given assembly descriptor to analyze.
     */
    public void mergeApplicationException(final AssemblyDescriptor assemblyDescriptor) {
        List<JApplicationException> assemblyDescriptorList = assemblyDescriptor.getApplicationExceptionList();
        if (assemblyDescriptorList != null) {
            for (JApplicationException applicationException : assemblyDescriptorList) {
                String className = applicationException.getClassName();
                // Get metadata for the given class
                ClassAnnotationMetadata exceptionMetadata = ejbMetadata.getClassAnnotationMetadata(encode(className));
                if (exceptionMetadata == null) {
                    throw new IllegalStateException("The class named '" + className
                            + "' was not found. Cannot set ApplicationException on it");
                }
                // Set the new value
                exceptionMetadata.setApplicationException(applicationException);
                logger.debug("Setting the application-exception ''{0}'' on the class ''{1}''", applicationException, className);
            }
        }
    }

    /**
     * Set/override the current interceptor-binding.
     * @param assemblyDescriptor the given assembly descriptor to analyze.
     */
    public void mergeInterceptorBinding(final AssemblyDescriptor assemblyDescriptor) {
        List<InterceptorBinding> interceptorBindingList = assemblyDescriptor.getInterceptorBindingList();
        if (interceptorBindingList != null) {
            for (InterceptorBinding interceptorBinding : interceptorBindingList) {

                // Get Ejb-Name
                String ejbName = interceptorBinding.getEjbName();
                // Default interceptor
                if (WILDCARD.equals(ejbName)) {
                    // Get current list
                    JInterceptors defaultInterceptors = ejbMetadata.getDefaultInterceptorsClasses();
                    if (defaultInterceptors == null) {
                        // build new list if empty
                        defaultInterceptors = new JInterceptors();
                        ejbMetadata.setDefaultInterceptorsClasses(defaultInterceptors);
                    }

                    // Add the interceptors of the interceptor-binding to the default interceptors
                    for (String interceptor : interceptorBinding.getInterceptorClassList()) {
                        defaultInterceptors.addClass(encode(interceptor));
                    }

                    // stop the loop
                    continue;
                }

                // Gets the metadata for the ejb
                ClassAnnotationMetadata beanMetadata = ejbMetadata.getClassAnnotationMetadataForEjbName(ejbName);
                if (beanMetadata == null) {
                    throw new IllegalStateException("No metadata found for the the ejb '" + ejbName
                            + "'. Check that this ejb is defined. This ejb-name is referenced from an interceptor-binding");
                }

                // It is a method ?
                MethodDD methodDD = interceptorBinding.getMethod();

                // not a method
                if (methodDD == null) {
                    JInterceptors interceptors = beanMetadata.getAnnotationInterceptors();
                    if (interceptors == null) {
                        // build new list if empty
                        interceptors = new JInterceptors();
                        beanMetadata.setAnnotationsInterceptors(interceptors);
                    }

                    for (String interceptor : interceptorBinding.getInterceptorClassList()) {
                        interceptors.addClass(encode(interceptor));
                    }

                    // Exclude default interceptors for this bean ?
                    if (interceptorBinding.isExcludeDefaultInterceptorsCalled()) {
                        beanMetadata.setExcludeDefaultInterceptors(interceptorBinding.isExcludeDefaultInterceptors());
                    }

                    // For ordered item: First, add interceptors (if they are not already present)
                    List<String> orderList = interceptorBinding.getOrderInterceptorClassList();
                    if (orderList != null) {

                        // Check that this is the first time that we try to order the list.
                        if (beanMetadata.isOrderedInterceptors()) {
                            throw new IllegalStateException("The interceptor-order element has already been used on the bean '"
                                    + ejbName + "'. It can only be used once.");
                        }

                        // Flag class as a class having ordered interceptors
                        beanMetadata.setOrderedInterceptors(true);

                        // Build a list with all interceptors. (it means that we
                        // add default interceptors if any) if they are not excluded
                        if (!beanMetadata.isExcludedDefaultInterceptors()) {
                            if (beanMetadata.getEjbJarAnnotationMetadata().getDefaultInterceptorsClasses() != null) {
                                for (String cls : beanMetadata.getEjbJarAnnotationMetadata().getDefaultInterceptorsClasses()
                                        .getClasses()) {
                                    interceptors.addClass(cls);
                                }
                            }
                        }

                        // Then add new items found in the DD
                        for (String interceptor : interceptorBinding.getOrderInterceptorClassList()) {
                            String encodedName = encode(interceptor);
                            if (!interceptors.contains(encodedName)) {
                                interceptors.addClass(encodedName);
                            }
                        }

                        // Order.
                        // First, check that the list has the correct size as it should be a total ordering.
                        if (interceptors.size() != orderList.size()) {
                            String err = "The list used for ordering interceptors is not a total list as the size mismatch.";
                            err += "The current list is '" + interceptors  + "' while the ordering list is '" + orderList + "'.";
                            throw new IllegalStateException(err);
                        }

                        // Size is ok, check that all elements are present
                        for (String clazz : orderList) {
                            if (!interceptors.contains(encode(clazz))) {
                                throw new IllegalStateException("The element '" + clazz
                                        + "' of the ordered list is not present in the interceptors '" + interceptors + "'.");
                            }
                        }

                        // Ok, all the elements are present, set the new ordering.
                        // Which is based on the given ordered list.
                        JInterceptors orderedInterceptors = new JInterceptors();
                        for (String clazz : orderList) {
                            orderedInterceptors.addClass(encode(clazz));
                        }
                        // Set the new list
                        beanMetadata.setAnnotationsInterceptors(orderedInterceptors);

                    }
                } else {
                    // method case
                    List<MethodAnnotationMetadata> methods = getMethodsForGivenMethodDD(methodDD, beanMetadata);

                    // If no elements, print warning
                    if (methods.size() == 0) {
                        logger.warn("Method with name '" + methodDD.getName()
                                + "' is specified in interceptorBinding element for class '" + beanMetadata.getClassName()
                                + "' but no methods are matching");
                    }

                    // now, apply interceptor on each method
                    for (MethodAnnotationMetadata method : methods) {
                        JInterceptors interceptors = method.getAnnotationInterceptors();
                        if (interceptors == null) {
                            // build new list if empty
                            interceptors = new JInterceptors();
                            method.setAnnotationsInterceptors(interceptors);
                        }

                        for (String interceptor : interceptorBinding.getInterceptorClassList()) {
                            interceptors.addClass(encode(interceptor));
                        }

                        // Exclude class interceptors for this method ?
                        if (interceptorBinding.isExcludeClassInterceptorsCalled()) {
                            method.setExcludeClassInterceptors(interceptorBinding.isExcludeClassInterceptors());
                        }
                        // Exclude default interceptors for this method ?
                        if (interceptorBinding.isExcludeDefaultInterceptorsCalled()) {
                            method.setExcludeDefaultInterceptors(interceptorBinding.isExcludeDefaultInterceptors());
                        }

                    }
                }
            }
        }
    }


    /**
     * Gets the methods metadata that are matching a given method DD.
     * @param methodDD the methodDD object
     * @param classAnnotationMetadata the class where to search methods.
     * @return the list of methods metadata
     */
    private static List<MethodAnnotationMetadata> getMethodsForGivenMethodDD(final MethodDD methodDD,
            final ClassAnnotationMetadata classAnnotationMetadata) {
        //get the method.
        String methodName = methodDD.getName();
        List<MethodAnnotationMetadata> methods = classAnnotationMetadata.searchMethodAnnotationMetadata(methodName);
        // params ?
        List<String> params = methodDD.getParams();
        // Keep only method with the correct params
        if (params != null && params.size() > 0) {
            List<MethodAnnotationMetadata> newMethods = new ArrayList<MethodAnnotationMetadata>();
            for (MethodAnnotationMetadata method : methods) {
                // get params for the current method
                Type[] types = Type.getArgumentTypes(method.getJMethod().getDescriptor());
                // Same size ?
                if (params.size() != types.length) {
                    // Not same size, go to the next one
                    continue;
                }

                // Compare each element
                int i = 0;
                boolean matching = false;
                for (String param : params) {
                    if (param.equals(types[i].getClassName())) {
                        matching = true;
                    }
                    i++;
                }
                // match ?
                if (matching) {
                    newMethods.add(method);
                }
            }
            // Get the new list (only matching elements)
            methods = newMethods;
        }
        return methods;
    }


    /**
     * Encode a classname (replace . by /) as it's the internal form of the
     * metadata
     * @param className the name of the class to encode
     * @return encoded name
     */
    private static String encode(final String className) {
        return className.replace(".", "/");
    }
}
